// -*- C++ -*-
/*!
 * @file DataOutPort.h
 * @brief Base class of OutPort
 * @date $Date: 2007/04/13 15:45:08 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: DataOutPort.h,v 1.7 2007/04/13 15:45:08 n-ando Exp $
 *
 */

/*
 * $Log: DataOutPort.h,v $
 * Revision 1.7  2007/04/13 15:45:08  n-ando
 * RTC::OK was changed to RTC::RTC_OK.
 *
 * Revision 1.6  2007/01/21 09:45:31  n-ando
 * To advertise interface profile, publishInterfaceProfile() is called now.
 *
 * Revision 1.5  2007/01/14 22:57:54  n-ando
 * A bug fix about template argument for buffer-type in constructor.
 *
 * Revision 1.4  2007/01/12 14:30:01  n-ando
 * A trivial bug fix.
 *
 * Revision 1.3  2007/01/06 17:44:11  n-ando
 * The behavior on notify_connect() and notify_disconnect() are now
 * implemented in protected functions(ex. publisherInterfaces()).
 *
 * Revision 1.2  2006/12/02 18:29:15  n-ando
 * Now OutPortCorbaProvider and InPortCorbaConsumer are used.
 *
 * Revision 1.1  2006/11/27 09:44:37  n-ando
 * The first commitment.
 *
 *
 */

#ifndef DataOutPort_h
#define DataOutPort_h


#include <rtm/PortBase.h>
#include <rtm/BufferBase.h>
#include <rtm/OutPortCorbaProvider.h>
#include <rtm/OutPortTcpSockProvider.h>
#include <rtm/InPortCorbaConsumer.h>
#include <rtm/InPortTcpSockConsumer.h>
#include <rtm/OutPort.h>
#include <rtm/NVUtil.h>
#include <rtm/PublisherFactory.h>
#include <rtm/Properties.h>

namespace RTC
{
  /*!
   * @if jp
   * @class DataOutPort
   * @brief InPort p Port
   * @else
   * @class DataOutPort
   * @brief InPort abstruct class
   * @endif
   */
  class DataOutPort
    : public PortBase
  {
  public:
    /*!
     * @if jp
     * @brief RXgN^
     * @else
     * @brief Constructor
     * @endif
     */
    template <class DataType, template <class DataType> class Buffer>
    DataOutPort(const char* name, OutPort<DataType, Buffer>& outport,
		Properties& prop)
      : PortBase(name), m_outport(outport)
    {
      // PortProfile::properties ݒ
      addProperty("port.port_type", "DataOutPort");
      
      m_providers.push_back(new OutPortCorbaProvider<DataType>(outport));
      m_providers.back()->publishInterfaceProfile(m_profile.properties);

      m_providers.push_back(new OutPortTcpSockProvider<DataType>(outport));
      m_providers.back()->publishInterfaceProfile(m_profile.properties);

      m_consumers.push_back(new InPortCorbaConsumer<DataType>(outport));
      m_consumers.push_back(new InPortTcpSockConsumer<DataType>(outport, prop));

    }


    /*!
     * @if jp
     * @brief fXgN^
     * @else
     * @brief Destructor
     * @endif
     */
    virtual ~DataOutPort();

    
    /*!
     * @if jp
     * @brief [CORBA interface] Port ̐ڑs
     *
     * OutPort  InPort Ƃ̐ڑsB
     *
     * OutPort  connect() ł͈ȉ̃V[PXŏsB
     *
     * 1. OutPort Ɋ֘A connector ̐уZbg
     *
     * 2. InPortɊ֘A connector ̎擾
     *  - ConnectorProfile::properties["dataport.corba_any.inport_ref"]
     *    OutPortAny ̃IuWFNgt@Xݒ肳ĂꍇA
     *    t@X擾ConsumerIuWFNgɃZbgB
     *    t@XZbgĂȂΖČpB
     *    (OutPortconnect() ĂяõGg|Cg̏ꍇ́A
     *    InPort̃IuWFNgt@X̓ZbgĂȂ͂łB)
     * 3. PortBase::connect() R[
     *    Port̐ڑ̊{sB
     * 4. L2.InPort̃t@X擾łȂ΁AēxInPort
     *    ֘A connector 擾B
     *
     * 5. ConnectorProfile::properties ŗ^ꂽ񂩂A
     *    OutPort̏sB
     *
     * - [dataport.interface_type]
     * -- CORBA_Any ̏ꍇ: 
     *    InPortAny ʂăf[^B
     *    ConnectorProfile::properties["dataport.corba_any.inport_ref"]
     *    InPortAny ̃IuWFNgt@XZbgB
     * -- RawTCP ̏ꍇ: Raw TCP socket ʂăf[^B
     *    ConnectorProfile::properties["dataport.raw_tcp.server_addr"]
     *    InPort̃T[oAhXZbgB
     *
     * - [dataport.dataflow_type]
     * -- Push̏ꍇ: Subscriber𐶐BSubscriber̃^Cv́A
     *    dataport.subscription_type ɐݒ肳ĂB
     * -- Pull̏ꍇ: InPortf[^Pull^Ŏ擾邽߁A
     *    ɉKvB
     *
     * - [dataport.subscription_type]
     * -- Oncȅꍇ: SubscriberOnce𐶐B
     * -- New̏ꍇ: SubscriberNew𐶐B
     * -- Periodic̏ꍇ: SubscriberPeriodic𐶐B
     *
     * - [dataport.push_interval]
     * -- dataport.subscription_type=Periodic̏ꍇݒ肷B
     *
     * 6. L̏̂łG[ł΁AG[^[B
     *    ɏsꂽꍇRTC::RTC_OKŃ^[B
     *  
     * @else
     * @brief [CORBA interface] Connect the Port
     * @endif
     */
    
  protected:
    /*!
     * @if jp
     *
     * @brief Interface J
     *
     * ̃Iy[V́Anotify_connect() V[PX̎n߂ɃR[
     * 鏃z֐łB
     * notify_connect() ł́A
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * ̏ protected ֐R[ڑsB
     * <br>
     * ̃Iy[V́AVK connector_id ɑ΂Ă͐ڑ̐A
     *  connector_id ɑ΂Ă͍XVK؂ɍsKvB
     *
     * @param connector_profile ڑɊւvt@C
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Publish interface information
     *
     * This operation is pure virutal method that would be called at the
     * beginning of the notify_connect() process sequence.
     * In the notify_connect(), the following methods would be called in order.
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile() 
     *
     * This operation should create the new connection for the new
     * connector_id, and should update the connection for the existing
     * connection_id.
     *
     * @param connector_profile The connection profile information
     * @return The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    publishInterfaces(ConnectorProfile& connector_profile);
    

    /*! @if jp
     *
     * @brief Interface 擾
     *
     * ̃Iy[V́Anotify_connect() V[PX̒ԂɃR[
     * 鏃z֐łB
     * notify_connect() ł́A
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * ̏ protected ֐R[ڑsB
     *
     * @param connector_profile ڑɊւvt@C
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Publish interface information
     *
     * This operation is pure virutal method that would be called at the
     * mid-flow of the notify_connect() process sequence.
     * In the notify_connect(), the following methods would be called in order.
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * @param connector_profile The connection profile information
     * @return The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    subscribeInterfaces(const ConnectorProfile& connector_profile);
    

    /*!
     * @if jp
     *
     * @brief Interface ̐ڑ
     *
     * ̃Iy[V́Anotify_disconnect() V[PX̏IɃR[
     * 鏃z֐łB
     * notify_disconnect() ł́A
     * - disconnectNext()
     * - unsubscribeInterfaces()
     * - eraseConnectorProfile()
     * ̏ protected ֐R[ڑsB
     *
     * @param connector_profile ڑɊւvt@C
     *
     * @else
     *
     * @brief Disconnect interface connection
     *
     * This operation is pure virutal method that would be called at the
     * end of the notify_disconnect() process sequence.
     * In the notify_disconnect(), the following methods would be called.
     * - disconnectNext()
     * - unsubscribeInterfaces()
     * - eraseConnectorProfile() 
     *
     * @param connector_profile The connection profile information
     *
     * @endif
     */
    virtual void
    unsubscribeInterfaces(const ConnectorProfile& connector_profile);


    
    struct publish
    {
      publish(SDOPackage::NVList& prop) : m_prop(prop) {}
      void operator()(OutPortProvider* provider)
      {
	provider->publishInterface(m_prop);
      }
      SDOPackage::NVList& m_prop;
    };

    struct unsubscribe
    {
      unsubscribe(const SDOPackage::NVList& prop) : m_prop(prop) {}
      void operator()(InPortConsumer* consumer)
      {
	consumer->unsubscribeInterface(m_prop);
      }
      const SDOPackage::NVList& m_prop;
    };

  private:
    std::vector<OutPortProvider*> m_providers;
    std::vector<InPortConsumer*> m_consumers;
    OutPortBase& m_outport;

    PublisherFactory m_pf;

    struct subscribe
    {
      subscribe(const ConnectorProfile& prof)
	: m_prof(&prof), _consumer(NULL) 
      {
      }

      subscribe(const subscribe& subs)
	: m_prof(subs.m_prof),
	  _consumer(subs._consumer)
      {
      }

      subscribe& operator=(const subscribe& subs)
      {
	if (this == &subs) return *this;
	m_prof = subs.m_prof;
	_consumer = subs._consumer;
	return *this;
      }

      void operator()(InPortConsumer* cons)
      {
	if (cons->subscribeInterface(m_prof->properties))
	  {
	    _consumer = cons;
	  }
      }
      const ConnectorProfile* m_prof;
      InPortConsumer* _consumer;
    };


  };
}; // namespace RTC

#endif // DataOutPort_h
