// -*- C++ -*-
/*!
 * @file StateMachine.h
 * @brief State machine template class
 * @date $Date: 2007/09/20 11:21:12 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: StateMachine.h,v 1.3.2.2 2007/09/20 11:21:12 n-ando Exp $
 *
 */

/*
 * $Log: StateMachine.h,v $
 * Revision 1.3.2.2  2007/09/20 11:21:12  n-ando
 * Some fixes.
 * - Template parameter "Result" is deleted.
 *   Now listener return type is void instead of the Result.
 * - Now exit action is invoked immediately, when the next state is changed in
 *   the entry action (do action is never invoked).
 *
 * Revision 1.3.2.1  2007/07/20 16:08:57  n-ando
 * A bug fix.
 *
 * Revision 1.3  2007/04/26 15:33:39  n-ando
 * The header include order was modified to define _REENTRANT before
 * including ace/config-lite.h in Linux systems.
 * In ace 5.4.7 or later, _REENTRANT flag should be defined explicitly.
 *
 * Revision 1.2  2007/01/09 15:26:53  n-ando
 * Now StateMachine does not need NOP function.
 * All null function pointer is checked and skipped.
 *
 * Revision 1.1  2006/10/26 08:55:53  n-ando
 * The first commitment.
 *
 *
 */

#ifndef StateMachine_h
#define StateMachine_h

#include <rtm/RTC.h>

#include <ace/Guard_T.h>
#include <ace/Thread_Mutex.h>

template <class State>
struct StateHolder
{
  State curr;
  State prev;
  State next;
};

/*!
 * @if jp
 *
 * @class StateMachine
 *
 * @brief ԃ}VNX
 *
 * StateMachine NX͏ԃ}VNXłB
 *
 * : ActiveObject͏ԃ}VANeBuIuWFNgłƂB
 * Ԃ3 INACTIVE, ACTIVE, ERROR AeԂłEntryExit
 * `ƂƁAȉ̂悤ɎB
 * <pre>
 * class ActiveObject 
 * {  
 * public: 
 *   enum MyState { INACTIVE, ACTIVE, ERROR }; 
 *   typedef States<MyState> MyStates; 
 *  
 *   ActiveObject() 
 *     : m_sm(3) 
 *   { 
 *     m_sm.setNOP(&ActiveObject::nullAction); 
 *     m_sm.setListener(this); 
 *  
 *     m_sm.setExitAction(NACTIVE, &ActiveObject::inactiveExit); 
 *       : 
 *     m_sm.setPostDoAction(ERROR, &ActiveObject::errorPostDo); 
 *     m_sm.setTransitionAction(&ActiveObject:tratransitionnsition); 
 *   }; 
 *  
 *   bool nullAction(MyStates st) {}; 
 *   bool inactiveExit(MyStates st) {}; 
 *     : 
 *   bool errorPostDo(MyStates st) {}; 
 *   bool transition(MyStates st) {}; 
 *  
 * private: 
 *   StateMachine<MyState, bool, ActiveObject> m_sm; 
 * }; 
 * </pre>
 * ԂNX͈ȉ̏𖞂悤ɎȂ΂ȂȂB
 * <ol>
 * <li> enum ŏԂ`
 * <li> StateMachine ̃ev[ǵA<br>
 *   <Ԃ̌^(MyState), ANV֐̖߂l(bool), YIuWFNǧ^>
 * <li> StateMachine ̃RXgN^͏Ԃ̐
 * <li> ȉ̃ANV֐(Return _function_name_(States)) ̊֐ƂĐݒ
 * <ol>
 *  <li> Ȃ֐K`AsetNOP ŗ^Ȃ΂ȂȂ
 *  <li> eԖ, set(Entry|PreDo|Do|PostDo|Exit)Action ŃANVݒ
 *  <li> ԑJڎ̃ANV setTransitionAction() ŐݒB
 * </ol>
 * <li> Jڎ̃ANV́A^ꂽݏԁAԁAOԂɁA
 *   [UȂ΂ȂȂB
 * <li> Ԃ̕ύX goTo() ŁAԂ̃`FbN isIn(state) ōsB
 * <li> goTo()͎ԂIɃZbg֐łAJڂ̉ۂ́A
 *   [UݏԂ擾f郍WbNȂ΂ȂȂB
 * </ol>
 *
 * ̃NX́A̏Ԃɑ΂āA
 * <ul>
 * <li> Entry action
 * <li> PreDo action
 * <li> Do action
 * <li> PostDo action
 * <li> Exit action
 * </ul>
 * 5̃ANV`邱ƂłB
 * Transition action ͂ԊԑJڂŌĂяoANVŁA
 * ̐U镑̓[U`Ȃ΂ȂȂB
 * 
 * ̃NX͈ȉ̂悤ȃ^C~OŊeANVsB
 *
 * <ul>
 * <li> ԂύX(A->B)ԂJڂꍇ <br>
 * (A:Exit)->|(ԍXV:A->B)->(B:Entry)->(B:PreDo)->(B:Do)->(B:PostDo)
 *
 * <li> ԂύXꂸABԂێꍇ (|̓Xebv̋؂\)<br>
 * (B(n-1):PostDo)->|(B(n):PreDo)->(B(n):Do)->(B(n):PostDo)->|(B(n+1):PreDo)<br>
 * PreDo, Do, PostDo JԂsB
 *
 * <li> ȑJڂꍇ <br>
 * (B(n-1):PostDo)->(B(n-1):Exit)->|(B(n):Entry)->(B(n):PreDo) <br>
 * U Exit Ă΂ꂽAEntry sAȍ~͑OƓB
 * </ul>
 * @else
 *
 * @brief
 *
 * @endif
 */
template <class State,
	  class Listener,
	  class States = StateHolder<State>, 
	  class Callback = void (Listener::*)(const States& states)
	  >
class StateMachine
{
public:
  /*!
   * @if jp
   * @brief RXgN^
   * @else
   * @brief Constructor
   * @endif
   */
  StateMachine(int num_of_state)
    : m_num(num_of_state),
      m_entry (new Callback[m_num]),
      m_predo (new Callback[m_num]),
      m_do    (new Callback[m_num]),
      m_postdo(new Callback[m_num]),
      m_exit  (new Callback[m_num])
  {
    setNullFunc(m_entry,  NULL);
    setNullFunc(m_do,     NULL);
    setNullFunc(m_exit,   NULL);
    setNullFunc(m_predo,  NULL);
    setNullFunc(m_postdo, NULL);
    m_transit = NULL;
  };


  /*!
   * @if jp
   * @brief NOP֐o^
   * @else
   * @brief Set NOP function
   * @endif
   */
  void setNOP(Callback call_back)
  {
    setNullFunc(m_entry,  call_back);
    setNullFunc(m_do,     call_back);
    setNullFunc(m_exit,   call_back);
    setNullFunc(m_predo,  call_back);
    setNullFunc(m_postdo, call_back);
    m_transit = call_back;
  }


  /*!
   * @if jp
   * @brief Listener IuWFNgo^
   * @else
   * @brief Set Listener Object
   * @endif
   */
  void setListener(Listener* listener)
  {
    m_listener = listener;
  }


  /*!
   * @if jp
   * @brief Entry action ֐o^
   * @else
   * @brief Set Entry action function
   * @endif
   */
  bool setEntryAction(State state, Callback call_back)
  {
    m_entry[state] = call_back;
    return true;
  }


  /*!
   * @if jp
   * @brief PreDo action ֐o^
   * @else
   * @brief Set PreDo action function
   * @endif
   */
  bool setPreDoAction(State state, Callback call_back)
  {
    m_predo[state] = call_back;
    return true;
  }


  /*!
   * @if jp
   * @brief Do action ֐o^
   * @else
   * @brief Set Do action function
   * @endif
   */
  bool setDoAction(State state, Callback call_back)
  {
    m_do[state] = call_back;
    return true;
    
  }


  /*!
   * @if jp
   * @brief Post action ֐o^
   * @else
   * @brief Set Post action function
   * @endif
   */
  bool setPostDoAction(State state, Callback call_back)
  {
    m_postdo[state] = call_back;
    return true;
  }


  /*!
   * @if jp
   * @brief Exit action ֐o^
   * @else
   * @brief Set Exit action function
   * @endif
   */
  bool setExitAction(State state, Callback call_back)
  {
    m_exit[state] = call_back;
    return true;
  }


  /*!
   * @if jp
   * @brief State transition action ֐o^
   * @else
   * @brief Set state transition action function
   * @endif
   */
  bool setTransitionAction(Callback call_back)
  {
    m_transit = call_back;
    return true;
  }


  /*!
   * @if jp
   * @brief ԂZbg
   * @else
   * @brief Set Exit action function
   * @endif
   */
  void setStartState(States states)
  {
    m_states.curr = states.curr;
    m_states.prev = states.prev;
    m_states.next = states.next;
  }


  /*!
   * @if jp
   * @brief Ԃ擾
   * @else
   * @brief Get state machine's status
   * @endif
   */
  States getStates()
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    return m_states;
  }
  
  State getState()
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    return m_states.curr;
  }
  

  /*!
   * @if jp
   * @brief ݏԂmF
   * @else
   * @brief Evaluate current status
   * @endif
   */
  bool isIn(State state)
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    return m_states.curr == state ? true : false;
  }


  /*!
   * @if jp
   * @brief ԂύX
   * @else
   * @brief Change status
   * @endif
   */
  void goTo(State state)
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    m_states.next = state;
    if (m_states.curr == state)
      {
	m_selftrans  = true;
      }
  }


  /*!
   * @if jp
   * @brief 쓮֐
   * @else
   * @brief Worker function
   * @endif
   */
  void worker()
  {
    States state;

    sync(state);

    if (state.curr == state.next)
      {
	// pre-do
	if (m_predo[state.curr] != NULL)
	  (m_listener->*m_predo [state.curr])(state);

	if (need_trans()) return;
	
	// do
	if (m_do[state.curr] != NULL)
	  (m_listener->*m_do    [state.curr])(state);

	if (need_trans()) return;
	
	// post-do
	if (m_postdo[state.curr] != NULL)
	  (m_listener->*m_postdo[state.curr])(state);
      }
    else
      {
	if (m_exit[state.curr] != NULL)
	  (m_listener->*m_exit[state.curr])(state);

	sync(state);
	
	if (state.curr != state.next)
	  {
	    state.curr = state.next;
	    if(m_entry[state.curr] != NULL)
	      (m_listener->*m_entry[state.curr])(state);
	    update_curr(state.curr);
	  }
      }
  }

protected:
  void setNullFunc(Callback* s, Callback nullfunc)
  {
    for (int i = 0; i < m_num; ++i) s[i] = nullfunc;
  }


  int m_num;
  Listener* m_listener;
  Callback* m_entry;
  Callback* m_predo;
  Callback* m_do;
  Callback* m_postdo;
  Callback* m_exit;
  Callback  m_transit;

  States m_states;
  bool m_selftrans;
  ACE_Thread_Mutex m_mutex;

private:
  inline void sync(States& st)
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    st = m_states;
  }
  
  inline bool need_trans()
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    return (m_states.curr != m_states.next);
  }

  inline void update_curr(const State curr)
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_mutex);
    m_states.curr = curr;
  }
};

#endif // StateMachine_h
